# PyTorch
import torch 
import torch.nn as nn 
import torch.nn.functional as F
# Utility
from CytoOne.utilities import ResidualBlock, reparameterize, kl_delta
# Typing 
from typing import Tuple, Optional


class Decoder(nn.Module):
    def __init__(self, 
                 input_dim: int,
                 batch_embedding_dim: int, 
                 latent_dims: list,
                 hidden_dims: list,
                 drop_out_p: float) -> None:
        """Initialize decoder

        Parameters
        ----------
        input_dim : int
            Dimension of the input 
        batch_embedding_dim : int
            Dimension of batch embedding 
        latent_dims : list
            A list of dimensions of latent variables 
        hidden_dims : list
            A nested list where each sublist contains the number of hidden units 
        drop_out_p : float
            Drop out probability
        """
        super().__init__()

        self.decoder_tower = nn.ModuleList()
        self.condition_z = nn.ModuleList()
        self.condition_xz = nn.ModuleList()
        current_d = latent_dims[0]*2
        for latent_d, hidden_d in zip(latent_dims[1:], hidden_dims[:-1]):
            self.decoder_tower.append(ResidualBlock(in_dim=current_d,
                                                    out_dim=latent_d,
                                                    hidden_dims=hidden_d,
                                                    drop_out_p=drop_out_p))
            # p(z_l | z_(l-1))
            self.condition_z.append(nn.Sequential(
                ResidualBlock(in_dim=latent_d, 
                              out_dim=latent_d,
                              hidden_dims=[latent_d,latent_d],
                              drop_out_p=drop_out_p),
                nn.GELU(),
                nn.Linear(latent_d, 2*latent_d)
            ))
            # p(z_l | x, z_(l-1))
            self.condition_xz.append(nn.Sequential(
                ResidualBlock(in_dim=latent_d*2,
                              out_dim=latent_d,
                              hidden_dims=[latent_d,latent_d],
                              drop_out_p=drop_out_p),
                nn.GELU(),
                nn.Linear(latent_d, 2*latent_d)
            ))
            current_d = latent_d*2
        
        
        self.recon = nn.Sequential(
            ResidualBlock(in_dim=current_d+batch_embedding_dim,
                            out_dim=current_d+batch_embedding_dim,
                            hidden_dims=hidden_dims[-1],
                            drop_out_p=drop_out_p),
            nn.Linear(current_d+batch_embedding_dim, 3*input_dim)
        )
        # Only used for pretrain model 
        self.decoder_elevator = nn.Sequential(
            nn.Linear(latent_dims[0]+batch_embedding_dim, 128),
            nn.LayerNorm(128),
            nn.GELU(),
            nn.Linear(128, 256),
            nn.LayerNorm(256),
            nn.GELU(),
            nn.Linear(256, 512),
            nn.LayerNorm(512),
            nn.GELU(),
            nn.Linear(512, 3*input_dim)
        )

    def forward(self, 
                z: torch.tensor, 
                batch_index: torch.tensor, 
                batch_embedding: torch.tensor,
                xs: Optional[list]=None, 
                mode: str="random",
                pretrain: bool=False) -> Tuple[torch.tensor, torch.tensor, torch.tensor, list, list]:
        """Forward for decoder 

        Parameters
        ----------
        z : torch.tensor
            Top level embedding 
        batch_index : torch.tensor
            Batch index 
        batch_embedding : torch.tensor
            All batch embeddings 
        xs : Optional[list], optional
            All letent embeddings, by default None
        mode : str, optional
            Whether to use random sampling, by default "random"
        pretrain : bool, optional
            Whether to use pretrain model, by default False

        Returns
        -------
        Tuple[torch.tensor, torch.tensor, torch.tensor, list, list]
            Mu, log-var, zero-inflation gate for reconstructed x. KL losses and all latent embeddings
        """
        b, w = z.shape
        batch_emb = batch_embedding(batch_index)
        zs = [z]
        if pretrain:
            direct_x_mu, direct_x_log_var, direct_x_gate_logit = self.decoder_elevator(torch.cat([z, batch_emb], dim=1)).chunk(3, dim=1)
            return direct_x_mu, direct_x_log_var, direct_x_gate_logit, [], zs
        else:
            decoder_out = torch.zeros(b, w, device=z.device, dtype=z.dtype)
            kl_losses = []
            for i in range(len(self.decoder_tower)):
                z_sample = torch.cat([decoder_out, z], dim=1)
                decoder_out = self.decoder_tower[i](z_sample)

                mu, log_var = self.condition_z[i](decoder_out).chunk(2, dim=1)

                if xs is not None:
                    delta_mu, delta_log_var = self.condition_xz[i](torch.cat([xs[i], decoder_out], dim=1)).chunk(2, dim=1)
                    kl_losses.append(kl_delta(delta_mu, delta_log_var, mu, log_var))
                    mu = mu + delta_mu
                    log_var = log_var + delta_log_var

                if mode == "fix":
                    z = reparameterize(mu, 0)
                else:
                    z = reparameterize(mu, torch.exp(0.5 * log_var))
                zs.append(z)

            decoder_out = torch.cat([decoder_out, z, batch_emb], dim=1)
            
            x_mu, x_log_var, x_gate_logit = self.recon(decoder_out).chunk(3, dim=1)
            return x_mu, x_log_var, x_gate_logit, kl_losses, zs

        

        


# Data IO
import os 
import json
# Data manipulation
import scanpy as sc 
import numpy as np 
import pandas as pd 
# Torch 
import torch 
import torch.nn as nn 
# Typing and other info 
from typing import Optional, Tuple, Union
from time import perf_counter
from contextlib import contextmanager
import psutil


@contextmanager
def process_time_ram(message: str=""):
    """Monitor time taken as well as memory usage

    Parameters
    ----------
    message : str, optional
        The message to be printed, by default ""
    """
    print("="*30)
    print(message)
    t1 = t2 = perf_counter()
    yield lambda: t2-t1
    t2 = perf_counter()
    print('RAM memory % used:', psutil.virtual_memory()[2])
    # Getting usage of virtual_memory in GB ( 4th field)
    print('RAM Used (GB):', psutil.virtual_memory()[3]/1000000000)
    print("Done. Time taken is {0:.4f} seconds.".format(t2-t1))
    print("="*30)


def import_data(cell_by_gene: Union[str, pd.DataFrame],
                cell_metadata: Optional[Union[str, pd.DataFrame]]=None,
                batch_index_col: Optional[str]=None,
                celltype_col: Optional[str]=None,
                normalize: bool=True,
                dr: bool=False,
                zero_inflated: bool=True) -> sc.AnnData:
    """Read in cell-by-gene matrix and meta information as an anndata object 

    Parameters
    ----------
    cell_by_gene : Union[str, pd.DataFrame]
        Either the path the .csv file or a pandas dataframe. We assume that 
        the first column is the cell ID  
    cell_metadata : Union[str, pd.DataFrame]
        Either the path the .csv file or a pandas dataframe. We assume that 
        the first column is the cell ID 
    batch_index_col : Optional[str], optional
        The column name in cell_metadata that contains the batch information for each cell, by default None
    celltype_col : Optional[str], optional
        The column name in cell_metadata that contains the cell type information for each cell, by default None
    normalize : bool, optional
        Whether or not to perform arcsinh-transformation, by default True
    dr : bool, optional
        Whether or not to perfrom UMAP, by default False
    zero_inflated : bool, optional
        Whether or not the data is zero-inflated, by default True

    Returns
    -------
    sc.AnnData
        An AnnData object containing curated information 

    """
    with process_time_ram("Processing cell-by-gene matrix") as ctm: 
        if isinstance(cell_by_gene, str) and ("csv" in os.path.splitext(cell_by_gene)[1]):
            counts = pd.read_csv(cell_by_gene, index_col=0)
        elif isinstance(cell_by_gene, pd.DataFrame):
            counts = cell_by_gene.copy()
        else: 
            raise TypeError("Only .csv file or pandas dataframe is allowed")
        
        counts.index.rename(name='cell_id', inplace=True)
        counts.index = counts.index.astype(str)
    
    with process_time_ram("Processing cell metadata") as ctm: 
        if cell_metadata is not None:
            if isinstance(cell_metadata, str) and ("csv" in os.path.splitext(cell_metadata)[1]):
                cell_meta = pd.read_csv(cell_metadata, index_col=0)
            elif isinstance(cell_metadata, pd.DataFrame):
                cell_meta = cell_metadata.copy()
            else: 
                raise TypeError("Only .csv file or pandas dataframe is allowed")
        else:
            batch_index_col = None
            cell_meta = pd.DataFrame({"batch": "0"}, index=counts.index)
        
        cell_meta.index.rename(name='cell_id', inplace=True)
        cell_meta.index = cell_meta.index.astype(str)

    
    with process_time_ram("Creating AnnData object") as ctm:
        adata = sc.AnnData(counts)
        # Add cell type information 
        if celltype_col is not None:
            adata.obs['cell_type'] = cell_meta.loc[adata.obs_names, celltype_col]
        else: 
            adata.obs['cell_type'] = "cell"
        # Create index for each cell type 
        adata.obs['leiden'] = pd.factorize(adata.obs['cell_type'])[0].astype(str)
        # Add batch information 
        if batch_index_col is None:
            adata.obs['batch'] = "0"
        else:  
            adata.obs['batch'] = cell_meta.loc[adata.obs_names, batch_index_col]
        
        adata.obs['batch_index'] = pd.factorize(adata.obs['batch'])[0].astype(int)
        # Add basic info
        adata.uns['unique_leiden'] = np.unique(adata.obs['leiden'])
        adata.uns['n_leiden'] = len(adata.uns['unique_leiden'])
        adata.uns['cell_type_leiden_map'] = adata.obs[["cell_type", "leiden"]].drop_duplicates(ignore_index=True)
        adata.uns['cell_type_leiden_map'].rename(columns={"cell_type": "cell_type_name",
                                                        "leiden": "cell_type_index"},
                                                inplace=True)
        
        adata.uns['unique_batch'] = np.unique(adata.obs['batch_index'])
        adata.uns['n_batches'] = len(adata.uns['unique_batch'])
        adata.uns['batch_index_batch_map'] = adata.obs[["batch", "batch_index"]].drop_duplicates(ignore_index=True)

        adata.uns['n_genes'] = adata.var.shape[0]
        adata.uns['zero_inflated'] = zero_inflated
    # If normalization is needed 
    if normalize:
        with process_time_ram("ArcSinh transform") as ctm:
            if isinstance(adata.X, np.ndarray):
                adata.X = np.arcsinh(adata.X / 5.0)
            else:
                # For sparse matrices
                adata.X = adata.X.toarray()
                adata.X = np.arcsinh(adata.X / 5.0)
    # If zero inflated, we make sure the lower bound is 0
    if zero_inflated:
        with process_time_ram("Clipping the data") as ctm:
            adata.X = np.clip(adata.X, a_min=0, a_max=None)
    # UMAP
    if dr:
        with process_time_ram("UMAP") as ctm:
            sc.pp.pca(adata)
            sc.pp.neighbors(adata)    
            sc.tl.umap(adata)

    return adata 


def generate_strata(adata: sc.AnnData, 
                    n_strata: int=100) -> pd.DataFrame:
    """Generate strata of the counts for sampling 
    This ensures that all batches are included in each minibatch 

    Parameters
    ----------
    adata : sc.AnnData
        An AnnData object 
    n_strata : int, optional
        The number of strata, by default 100

    Returns
    -------
    pd.DataFrame
        A pd.DataFrame with batch information and strata information 
    """
    adata_w_batch = adata.to_df().copy()
    adata_w_batch['batch_index'] = adata.obs['batch_index'].copy()
    # Randomly suffle each batch 
    adata_w_batch = adata_w_batch.groupby(['batch_index'], observed=True, as_index=True).apply(lambda x: x.sample(frac=1, replace=False),
                                                            include_groups=False)
    adata_w_batch.reset_index(drop=False, names=['batch_index', "id_to_drop"], inplace=True)
    adata_w_batch.drop(columns=['id_to_drop'], inplace=True)
    # Split each batch into n_strata
    split_result = []
    for batch, group in adata_w_batch.groupby("batch_index"):
        splits = np.array_split(group.index, n_strata)
        for i, row_ind in enumerate(splits):
            part = group.loc[row_ind, :].copy()
            part["batch_index"] = batch
            part["stratum"] = i
            split_result.append(part)
    # Concatenate the results 
    adata_w_batch_strata = pd.concat(split_result).reset_index(drop=True)
    return adata_w_batch_strata


def load_stratum(adata_w_batch_strata: pd.DataFrame, 
                 stratum_id: int, 
                 target_batch_index: Optional[int]=None,
                 model_device='cpu') -> Tuple[torch.tensor, torch.tensor, torch.tensor]:
    """Load a stratum given its id 

    Parameters
    ----------
    adata_w_batch_strata : pd.DataFrame
        A dataframe with batch and strata information 
    stratum_id : int
        The id for the stratum to load 
    target_batch_index : Optional[int], optional
        If None, it would default to the current batch index. If not, it can be used for batch correction, by default None
    model_device : str, optional
        The model device , by default 'cpu'

    Returns
    -------
    Tuple[torch.tensor, torch.tensor, torch.tensor]
        Tensor for cell-by-gene, source batch index, target batch index 
    """
    # Extract the patch 
    # Other than the protein channels it will also contain stratum and batch 
    cell_patch = adata_w_batch_strata.loc[adata_w_batch_strata['stratum'] == stratum_id, :].copy()
    # Convert batch to tensor
    source_batch_index = torch.tensor(cell_patch['batch_index'].values, dtype=torch.int64, device=model_device)
    if target_batch_index is None:
        target_batch_index = source_batch_index.clone()
    else:
        # Useful for batch correction 
        target_batch_index = torch.ones_like(source_batch_index)*target_batch_index
    # Only keep counts 
    cell_patch.drop(columns=['stratum', 'batch_index'], inplace=True)
    cell_by_gene_counts = torch.tensor(cell_patch.values, dtype=torch.float32, device=model_device)
    
    return cell_by_gene_counts, source_batch_index, target_batch_index

    
class ResidualBlock(nn.Module):
    def __init__(self, in_dim: int, 
                 out_dim: int, 
                 hidden_dims: list,
                 drop_out_p: float) -> None:
        """Residual block used in encoder and decoder

        Parameters
        ----------
        in_dim : int
            Dimension of the input 
        out_dim : int
            Dimension of the output 
        hidden_dims : list
            A nested list describing the number of hidden units 
        drop_out_p : float
            Probability of drop out 
        """
        super().__init__()

        self.skip_proj = nn.Linear(in_dim, out_dim) if in_dim != out_dim else nn.Identity()
        
        self.residual_seq = nn.ModuleList()
        current_d = in_dim
        for h_d in hidden_dims:
            self.residual_seq.append(nn.Sequential(
                nn.Linear(current_d, h_d),
                nn.LayerNorm(h_d),
                nn.GELU(),
                nn.Dropout(drop_out_p)
            ))
            current_d = h_d       
        self.residual_seq.append(nn.Sequential(
                    nn.Linear(current_d, out_dim),
                    nn.LayerNorm(out_dim)
                ))  
    def forward(self, x):
        skip_x = self.skip_proj(x)
        for s in self.residual_seq:
            x = s(x)
        return skip_x + 0.1*x


def kl_delta(delta_mu: torch.tensor,
             delta_log_var: torch.tensor,
             mu: torch.tensor,
             log_var: torch.tensor) -> torch.tensor:
    """The KL for all other z's than the very top one 
        The output of the encoder is configured to output delta changes 
    Parameters
    ----------
    delta_mu : torch.tensor
        Delta of the mu
    delta_log_var : torch.tensor
        Delta of the log variance 
    mu : torch.tensor
        Base mu 
    log_var : torch.tensor
        Base log variance 

    Returns
    -------
    torch.tensor
        The KL divergence 
    """
    var = torch.exp(log_var)
    delta_var = torch.exp(delta_log_var)
    loss = -0.5 * torch.sum(1 + delta_log_var - delta_mu ** 2 / var - delta_var, dim=[1])
    return torch.mean(loss, dim=0)


def kl_standard(mu: torch.tensor,
                log_var: torch.tensor) -> torch.tensor:
    """KL divergence between Isotropic normal 

    Parameters
    ----------
    mu : torch.tensor
        Base mu
    log_var : torch.tensor
        Base log variance 

    Returns
    -------
    torch.tensor
        The KL divergence 
    """
    # KL(q||p) p=N(0,1) \int q log q/p
    loss = -0.5 * torch.sum(1 + log_var - mu ** 2 - torch.exp(log_var), dim=[1])
    return torch.mean(loss, dim=0)


def reparameterize(mu: torch.tensor,
                   std: torch.tensor) -> torch.tensor:
    """Sample normal using the reparameterization trick

    Parameters
    ----------
    mu : torch.tensor
        Mean
    std : torch.tensor
        Standard deviation 

    Returns
    -------
    torch.tensor
        Normal sample 
    """
    z = torch.randn_like(mu) * std + mu
    return z


def compute_mmd(x: torch.tensor,
                y: torch.tensor) -> torch.tensor:
    """Compute maximum mean discrapency between two tensors 

    Parameters
    ----------
    x : torch.tensor
        An N*p tensor
    y : torch.tensor
        An M*p tensor 

    Returns
    -------
    torch.tensor
        MMD with RBF kernel 
    """
    xx = torch.mm(x, x.t())
    yy = torch.mm(y, y.t())
    xy = torch.mm(x, y.t())
    
    # RBF Kernel
    bandwidth_range = [0.01, 0.1, 1, 10]
    # k(x,x)
    xx_sum = torch.diag(xx).unsqueeze(1).expand(xx.size(0), xx.size(0))
    xx_rbf = xx_sum + xx_sum.t() - 2*xx
    # k(y,y)
    yy_sum = torch.diag(yy).unsqueeze(1).expand(yy.size(0), yy.size(0))
    yy_rbf = yy_sum + yy_sum.t() - 2*yy
    # k(x,y)
    xy_sum_1 = torch.diag(xx).unsqueeze(1).expand(xx.size(0), yy.size(0))
    xy_sum_2 = torch.diag(yy).unsqueeze(0).expand(xx.size(0), yy.size(0))
    xy_rbf = xy_sum_1 + xy_sum_2 - 2*xy
    # Sum over differnt bandwidthes
    mmd = 0
    for bandwidth in bandwidth_range:
        xx_kernel = torch.exp(-xx_rbf / (2 * bandwidth ** 2))
        yy_kernel = torch.exp(-yy_rbf / (2 * bandwidth ** 2))
        xy_kernel = torch.exp(-xy_rbf / (2 * bandwidth ** 2))
        
        mmd += torch.mean(xx_kernel) + torch.mean(yy_kernel) - 2 * torch.mean(xy_kernel)
            
    return mmd


def mmd_loss(z: torch.tensor,
             batch_index: torch.tensor) -> torch.tensor:
    """Compute MMD between all pairs of batches 

    Parameters
    ----------
    z : torch.tensor
        The latent embedding 
    batch_index : torch.tensor
        The corresponding batch indices 

    Returns
    -------
    torch.tensor
        MMD 
    """
    unique_batches = batch_index.unique()
    mmd_sum = 0
    count = 0
    # For each unique pair
    for i in range(len(unique_batches)):
        for j in range(i + 1, len(unique_batches)):
            idx_i = (batch_index == unique_batches[i])
            idx_j = (batch_index == unique_batches[j])
            # Extract embedding 
            zi = z[idx_i]
            zj = z[idx_j]
            # Compute MMD 
            mmd = compute_mmd(x=zi,
                              y=zj)
            
            mmd_sum += mmd
            count += 1
    return mmd_sum / count if count > 0 else 0

class JSONEncoder(json.JSONEncoder):
    """This class is used to save dictionary of pd.DataFrame to a json file

    """
    def default(self, obj):
        if hasattr(obj, 'to_json'):
            return obj.to_json(orient='records')
        return json.JSONEncoder.default(self, obj)


# PyTorch
import torch 
import torch.nn as nn 
import torch.nn.functional as F
# Utility
from CytoOne.utilities import ResidualBlock
# Typing 
from typing import Tuple

class Encoder(nn.Module):
    def __init__(self,
                 input_dim: int,
                 batch_embedding_dim: int, 
                 latent_dims: list,
                 hidden_dims: list,
                 drop_out_p: float) -> None:
        """Initialize encoder

        Parameters
        ----------
        input_dim : int
            Dimension of the input 
        batch_embedding_dim : int
            Dimension of batch embedding 
        latent_dims : list
            A list of dimensions of latent variables 
        hidden_dims : list
            A nested list where each sublist contains the number of hidden units 
        drop_out_p : float
            Probability of drop out 
        """
        super().__init__()

        # The encoder module takes x and batch embedding 
        # (x + batch) -> latent_dims[0] -> latent_dims[1] -> ...
        self.encoder_tower = nn.ModuleList()
        current_d = input_dim+batch_embedding_dim
        for latent_d, hidden_d in zip(latent_dims, hidden_dims):
            self.encoder_tower.append(ResidualBlock(in_dim=current_d,
                                                    out_dim=latent_d,
                                                    hidden_dims=hidden_d,
                                                    drop_out_p=drop_out_p))
            current_d = latent_d
        
        # Generate mu and log_var for the top-level z 
        self.condition_x = nn.Sequential(
            nn.GELU(),
            nn.Linear(current_d, 2*current_d)
        )
        # This module directly mapps from X to top-level z
        self.encoder_elevator = nn.Sequential(
            nn.Linear(input_dim+batch_embedding_dim, 512),
            nn.LayerNorm(512),
            nn.GELU(),
            nn.Linear(512, 256),
            nn.LayerNorm(256),
            nn.GELU(),
            nn.Linear(256, 128),
            nn.LayerNorm(128),
            nn.GELU(),
            nn.Linear(128, 2*current_d)
        )

    def forward(self, 
                x: torch.tensor, 
                batch_index: torch.tensor, 
                batch_embedding: torch.tensor, 
                pretrain: bool) -> Tuple[torch.tensor, torch.tensor, list]:
        """Forward for encoder

        Parameters
        ----------
        x : torch.tensor
            Input 
        batch_index : torch.tensor
            Batch index 
        batch_embedding : torch.tensor
            All batch embeddings 
        pretrain : bool
            If use pretrain model 

        Returns
        -------
        Tuple[torch.tensor, torch.tensor, list]
            Top level mu, log_var, and all other embeddings 
        """
        batch_emb = batch_embedding(batch_index)
        x = torch.cat([x, batch_emb], dim=1)
        direct_mu, direct_log_var = self.encoder_elevator(x).chunk(2, dim=1)
        if pretrain:
            return  direct_mu, direct_log_var, []
        else:
            xs = []
            last_x = x
            for e in self.encoder_tower:
                x = e(x)
                last_x = x
                xs.append(x)

            mu, log_var = self.condition_x(last_x).chunk(2, dim=1)
            # xs is now [latent_dims[0], latent_dims[1], ...]
            # we do not need the top-level for the decoder
            # To make indexing a litter easier, we also reverse the order 
            return direct_mu + 0.1*mu, direct_log_var + 0.1*log_var, xs[:-1][::-1] 



import pytest
import numpy as np
import pandas as pd 
from CytoOne.cytoone_class import cytoone

counts_zi = pd.read_csv("./tests/test_data_zi.csv", index_col=0)
counts_zi.index.rename(name='cell_id', inplace=True)
counts_n = pd.read_csv("./tests/test_data_n.csv", index_col=0)
counts_n.index.rename(name='cell_id', inplace=True)
meta = pd.read_csv("./tests/test_data_meta.csv", index_col=0)
meta.index.rename(name='cell_id', inplace=True)


@pytest.mark.parametrize("batch_index_col, normalize, dr, zero_inflated, distribution_type, cell_by_gene, cell_metadata, target_batch_index, get_normal_component",
                         [
                             ("batch", True, True, True, "softplus_normal", counts_zi, meta, None, False),
                             ("batch", True, True, True, "log_normal", counts_zi, meta, None, False),
                             ("batch", True, True, True, "normal", counts_zi, meta, None, False),
                         ])
def test_cytoone(batch_index_col,
                 normalize,
                 dr,
                 zero_inflated,
                 distribution_type,
                 cell_by_gene,
                 cell_metadata,
                 target_batch_index,
                 get_normal_component):
    cyto = cytoone(batch_index_col=batch_index_col,
                   celltype_col="cell_type",
                   normalize=normalize,
                   dr=dr,
                   zero_inflated=zero_inflated,
                   distribution_type=distribution_type) 
    cyto.import_data(cell_by_gene=cell_by_gene,
                     cell_metadata=cell_metadata)
    cyto.initialize_parameters()
    cyto.training_loop(n_epoches=1)
    x_samples, z_samples = cyto.infer(target_batch_index=target_batch_index,
                                        get_normal_component=get_normal_component)
    assert (x_samples.shape[0]==4000) and (z_samples.shape[1]==2)
    
    



"""The Command-Line Interface (CLI) of CytoOne

The CLI of CytoOne can be accessed via ``python -m CytoOne``.

:Example:

    Get help:
    
    .. code-block:: bash

        python -m CytoOne -h
    
    Check version and authors:
    
    .. code-block:: bash
    
        python -m CytoOne --version 
        python -m CytoOne --author

**Pros**
    * Shallow learning curve
    
**Cons**
    * Fewer configurations 
    * No inspections of intermediate results
"""


import os
import sys
import torch
import argparse

from CytoOne.__version__ import __version__, __author__
from CytoOne.cytoone_class import cytoone


parser = argparse.ArgumentParser(description="CytoOne")

parser.add_argument("--version", action="version",
                    version=__version__, help="Display the version of the software")
parser.add_argument("--author", action="version", version=__author__,
                    help="Check the author list of the algorithm")
# Instantiate object 
parser.add_argument("--batch_index_col", nargs="?", type=str,
                    help="The column containing batch information")
parser.add_argument("--celltype_col", nargs="?", type=str,
                    help="The column cell type information")
parser.add_argument("--normalize", action="store_true",
                    help="If normalize the data via asinh")
parser.add_argument("--zero_inflated", action="store_true",
                    help="If the data if zero-inflated")
# Import data 
parser.add_argument("--cell_by_gene", type=str,
                    help="The path to the csv file that contains the CyTOF measurements")
parser.add_argument("--cell_metadata", type=str,
                    help="The path to the csv file that contains metainformation of the CyTOF measurements")
# Model training
parser.add_argument("--n_epoches", type=int, default=50,
                    help="Number of epoches")
parser.add_argument("--n_strata", type=int, default=100,
                    help="Number of minibatches per epoch")
# Downstream analysis 
parser.add_argument("--target_batch_index", type=int, nargs="?",
                    help="Batch to which batch correction is targeted")
parser.add_argument("--get_normal_component", action='store_true',
                    help="Whether to get the normal component")
# Model saving 
parser.add_argument("--dir_name", type=str, default=".",
                    help="The directory path at which the saved model should be.")
parser.add_argument("--model_name", type=str, default="cytoone",
                    help="The model name")



def main(cmdargs: argparse.Namespace):
    """The main method for CytoOne

    Parameters:
    ----------
    cmdargs: argparse.Namespace
        The command line argments and flags 
    """
    batch_index_col = cmdargs.batch_index_col
    celltype_col = cmdargs.celltype_col
    normalize = cmdargs.normalize
    zero_inflated = cmdargs.zero_inflated

    cyto = cytoone(batch_index_col=batch_index_col,
                   celltype_col=celltype_col,
                   normalize=normalize,
                   dr=False,
                   zero_inflated=zero_inflated)
    
    cell_by_gene = cmdargs.cell_by_gene 
    cell_metadata = cmdargs.cell_metadata
    
    cyto.import_data(cell_by_gene=cell_by_gene,
                     cell_metadata=cell_metadata)
    
    cyto.initialize_parameters()

    n_epoches = cmdargs.n_epoches
    n_strata = cmdargs.n_strata

    cyto.training_loop(n_epoches=n_epoches,
                       n_strata=n_strata)
    
    target_batch_index = cmdargs.target_batch_index
    get_normal_component = cmdargs.get_normal_component

    x_samples, z_samples = cyto.infer(target_batch_index=target_batch_index,
                                        get_normal_component=get_normal_component)

    dir_name = cmdargs.dir_name 
    model_name = cmdargs.model_name 
    cyto.save_model(dir_name=dir_name,
                    model_name=model_name)
    x_samples.to_csv(os.path.join(dir_name, model_name+"_x_samples.csv"),
                     index=True)
    z_samples.to_csv(os.path.join(dir_name, model_name+"_z_samples.csv"),
                     index=True)

    sys.exit(0)


if __name__ == "__main__":
    cmdargs = parser.parse_args()
    main(cmdargs=cmdargs)


# Quadrature 
import math
import numpy as np 
from numpy.polynomial.hermite import hermgauss 
# PyTorch
import torch
import torch.nn as nn 
from torch.distributions import constraints
from torch.distributions.utils import (
    broadcast_all,
    lazy_property,
    logits_to_probs,
    probs_to_logits,
)
from torch.nn.functional import softplus
from torch.distributions.normal import Normal
from torch.distributions.transformed_distribution import TransformedDistribution
from torch.distributions.transforms import SoftplusTransform, ExpTransform
# Pyro 
from pyro.distributions import TorchDistribution
from pyro.distributions.util import broadcast_shape


class SoftplusNormal(TransformedDistribution):
    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.positive
    has_rsample = True

    def __init__(self, loc, scale, validate_args=None):
        base_dist = Normal(loc, scale, validate_args=validate_args)
        super().__init__(base_dist, SoftplusTransform(), validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(SoftplusNormal, _instance)
        return super().expand(batch_shape, _instance=new)


class LogNormal(TransformedDistribution):
    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.positive
    has_rsample = True

    def __init__(self, loc, scale, validate_args=None):
        base_dist = Normal(loc, scale, validate_args=validate_args)
        super().__init__(base_dist, ExpTransform(), validate_args=validate_args)

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(LogNormal, _instance)
        return super().expand(batch_shape, _instance=new)
    

class QuasiZeroInflatedPositiveDistribution(TorchDistribution):

    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
        "normal_scale": constraints.greater_than_eq(0),
    }

    def __init__(self, base_dist, *, gate=None, gate_logits=None, 
                 normal_scale=None, quadrature_degree=20, validate_args=None):
        if (gate is None) == (gate_logits is None):
            raise ValueError(
                "Either `gate` or `gate_logits` must be specified, but not both."
            )
        if gate is not None:
            batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
            self.gate = gate.expand(batch_shape)
        else:
            batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
            self.gate_logits = gate_logits.expand(batch_shape)
        if base_dist.event_shape:
            raise ValueError(
                "QuasiZeroInflatedDistribution expected empty "
                "base_dist.event_shape but got {}".format(base_dist.event_shape)
            )
        if (normal_scale is None) or (normal_scale <= 0):
            self.normal_scale = torch.zeros(1)
            self.is_quasi = False
        else:
            self.normal_scale = normal_scale
            self.is_quasi = True
        # Gauss-Hermite nodes and weights (numpy arrays)
        gh_x, gh_w = hermgauss(quadrature_degree)
        self.gh_x = torch.tensor(gh_x, dtype=torch.float32)  # shape [n]
        self.gh_w = torch.tensor(gh_w, dtype=torch.float32)  # shape [n]
        self.base_dist = base_dist.expand(batch_shape)
        event_shape = torch.Size()

        super().__init__(batch_shape, event_shape, validate_args)

    @constraints.dependent_property
    def support(self):
        return self.base_dist.support

    @lazy_property
    def gate(self):
        return logits_to_probs(self.gate_logits, is_binary=True)

    @lazy_property
    def gate_logits(self):
        return probs_to_logits(self.gate, is_binary=True)
    
    def zero_normal_log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        # compute the variance
        var = self.normal_scale**2
        log_scale = (
            math.log(self.normal_scale)
            if isinstance(self.normal_scale, float)
            else self.normal_scale.log()
        )
        return (
            -((value) ** 2) / (2 * var)
            - log_scale
            - math.log(math.sqrt(2 * math.pi))
        ) 

    def log_prob(self, value):
        if self._validate_args:
            self._validate_sample(value)
        if not self.is_quasi:
            if "gate" in self.__dict__:
                gate, value = broadcast_all(self.gate, value)
                log_prob = torch.where(value <= 0, (gate).log(), (-gate).log1p() + self.base_dist.log_prob(value+1e-7))
            else:
                gate_logits, value = broadcast_all(self.gate_logits, value)
                log_prob = torch.where(value <= 0, 
                                    gate_logits-softplus(gate_logits), 
                                    -gate_logits + self.base_dist.log_prob(value+1e-7)-softplus(-gate_logits))
        else:
            if "gate" in self.__dict__:
                gate, normal_scale, value = broadcast_all(self.gate, self.normal_scale, value)
                
                gh_x = self.gh_x.view((-1,) + (1,) * value.ndim)  # [n, ...]
                gh_w = self.gh_w.view((-1,) + (1,) * (value.ndim))  # [n, ...]
                # If bernoulli is 0 (p) the prob is that of a normal
                zero_p = (gate.log() + self.zero_normal_log_prob(value=value)).exp()
                # If bernoulli is 1 (1-p), gauss hermite with extended support 
                # for n*p value, this will give us 
                # d_quadrature * n * p tensor
                temp_x = gh_x * math.sqrt(2) * normal_scale + value
                x = torch.where(temp_x>1e-7, temp_x, 1e-9)
                w = torch.where(temp_x>1e-7, gh_w, 0)
                gh_int = (self.base_dist.log_prob(x).exp()*w).sum(dim=0)/math.sqrt(math.pi)
                one_p = (-gate).log1p() + (gh_int+1e-7).log()
                log_prob = (zero_p.exp() + one_p.exp()).log()
            else:
                gate_logits, normal_scale, value = broadcast_all(self.gate_logits, self.normal_scale, value)
                
                gh_x = self.gh_x.view((-1,) + (1,) * value.ndim)  # [n, ...]
                gh_w = self.gh_w.view((-1,) + (1,) * (value.ndim))  # [n, ...]
                # If bernoulli is 0 (p) the prob is that of a normal
                zero_p = gate_logits-softplus(gate_logits) + self.zero_normal_log_prob(value=value)
                # If bernoulli is 1 (1-p), gauss hermite with extended support 
                # for n*p value, this will give us 
                # d_quadrature * n * p tensor
                temp_x = gh_x * math.sqrt(2) * normal_scale + value
                x = torch.where(temp_x>1e-7, temp_x, 1e-9)
                w = torch.where(temp_x>1e-7, gh_w, 0)
                gh_int = (self.base_dist.log_prob(x).exp()*w).sum(dim=0)/math.sqrt(math.pi)
                one_p = -gate_logits -softplus(-gate_logits) + (gh_int+1e-7).log()
                log_prob = (zero_p.exp() + one_p.exp()).log() 
        return log_prob

    def sample(self, sample_shape=torch.Size()):
        shape = self._extended_shape(sample_shape)
        with torch.no_grad():
            mask = torch.bernoulli(self.gate.expand(shape)).bool()
            samples = self.base_dist.expand(shape).sample()
            samples = torch.where(mask, samples.new_zeros(()), samples)
            noise = 0.0
            if self.is_quasi:
                noise = torch.normal(0.0, self.normal_scale.expand(shape))
        return samples + noise

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(type(self), _instance)
        batch_shape = torch.Size(batch_shape)
        gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
        gate_logits = (
            self.gate_logits.expand(batch_shape)
            if "gate_logits" in self.__dict__
            else None
        )
        normal_scale = self.normal_scale.expand(batch_shape)
        base_dist = self.base_dist.expand(batch_shape)
        QuasiZeroInflatedPositiveDistribution.__init__(
            new, base_dist, gate=gate, gate_logits=gate_logits,\
            normal_scale=normal_scale, quadrature_degree=len(self.gh_x), \
            validate_args=False
        )
        new._validate_args = self._validate_args
        return new


class QuasiZeroInflatedSoftplusNormal(QuasiZeroInflatedPositiveDistribution):
    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
        "normal_scale": constraints.greater_than_eq(0),
    }
    support = constraints.real

    def __init__(self, loc, scale, *, 
                 gate=None, gate_logits=None, 
                 normal_scale=None, quadrature_degree=20,
                 validate_args=None):
        base_dist = SoftplusNormal(loc=loc, 
                          scale=scale, validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits,
            normal_scale=normal_scale, quadrature_degree=quadrature_degree,
            validate_args=validate_args
        )


class QuasiZeroInflatedLogNormal(QuasiZeroInflatedPositiveDistribution):
    arg_constraints = {
        "gate": constraints.unit_interval,
        "gate_logits": constraints.real,
        "normal_scale": constraints.greater_than_eq(0),
    }
    support = constraints.real

    def __init__(self, loc, scale, *, 
                 gate=None, gate_logits=None, 
                 normal_scale=None, quadrature_degree=20,
                 validate_args=None):
        base_dist = LogNormal(loc=loc, 
                          scale=scale, validate_args=False)
        base_dist._validate_args = validate_args
        super().__init__(
            base_dist, gate=gate, gate_logits=gate_logits,
            normal_scale=normal_scale, quadrature_degree=quadrature_degree,
            validate_args=validate_args
        )
        


# class ZeroInflatedPositiveDistribution(TorchDistribution):

#     arg_constraints = {
#         "gate": constraints.unit_interval,
#         "gate_logits": constraints.real,
#     }

#     def __init__(self, base_dist, *, gate=None, gate_logits=None, validate_args=None):
#         if (gate is None) == (gate_logits is None):
#             raise ValueError(
#                 "Either `gate` or `gate_logits` must be specified, but not both."
#             )
#         if gate is not None:
#             batch_shape = broadcast_shape(gate.shape, base_dist.batch_shape)
#             self.gate = gate.expand(batch_shape)
#         else:
#             batch_shape = broadcast_shape(gate_logits.shape, base_dist.batch_shape)
#             self.gate_logits = gate_logits.expand(batch_shape)
#         if base_dist.event_shape:
#             raise ValueError(
#                 "ZeroInflatedDistribution expected empty "
#                 "base_dist.event_shape but got {}".format(base_dist.event_shape)
#             )

#         self.base_dist = base_dist.expand(batch_shape)
#         event_shape = torch.Size()

#         super().__init__(batch_shape, event_shape, validate_args)

#     @constraints.dependent_property
#     def support(self):
#         return self.base_dist.support

#     @lazy_property
#     def gate(self):
#         return logits_to_probs(self.gate_logits, is_binary=True)

#     @lazy_property
#     def gate_logits(self):
#         return probs_to_logits(self.gate, is_binary=True)

#     def log_prob(self, value):
#         if self._validate_args:
#             self._validate_sample(value)

#         if "gate" in self.__dict__:
#             gate, value = broadcast_all(self.gate, value)
#             log_prob = torch.where(value == 0, (gate).log(), (-gate).log1p() + self.base_dist.log_prob(value))
#         else:
#             gate_logits, value = broadcast_all(self.gate_logits, value)
#             log_prob = torch.where(value == 0, 
#                                    gate_logits-softplus(gate_logits), 
#                                    -gate_logits + self.base_dist.log_prob(value+1e-7)-softplus(-gate_logits))
#         return log_prob

#     def sample(self, sample_shape=torch.Size()):
#         shape = self._extended_shape(sample_shape)
#         with torch.no_grad():
#             mask = torch.bernoulli(self.gate.expand(shape)).bool()
#             samples = self.base_dist.expand(shape).sample()
#             samples = torch.where(mask, samples.new_zeros(()), samples)
#         return samples

#     def expand(self, batch_shape, _instance=None):
#         new = self._get_checked_instance(type(self), _instance)
#         batch_shape = torch.Size(batch_shape)
#         gate = self.gate.expand(batch_shape) if "gate" in self.__dict__ else None
#         gate_logits = (
#             self.gate_logits.expand(batch_shape)
#             if "gate_logits" in self.__dict__
#             else None
#         )
#         base_dist = self.base_dist.expand(batch_shape)
#         ZeroInflatedPositiveDistribution.__init__(
#             new, base_dist, gate=gate, gate_logits=gate_logits, validate_args=False
#         )
#         new._validate_args = self._validate_args
#         return new


# class ZeroInflatedSoftplusNormal(ZeroInflatedPositiveDistribution):
#     arg_constraints = {
#         "gate": constraints.unit_interval,
#         "gate_logits": constraints.real,
#     }
#     support = constraints.greater_than_eq(0)

#     def __init__(self, loc, scale, *, 
#                  gate=None, gate_logits=None, validate_args=None):
#         base_dist = SoftplusNormal(loc=loc, 
#                           scale=scale, validate_args=False)
#         base_dist._validate_args = validate_args

#         super().__init__(
#             base_dist, gate=gate, gate_logits=gate_logits, validate_args=validate_args
#         )



from CytoOne import __main__
import CytoOne
import pytest
import numpy as np

import sys
import os
import shutil
from io import StringIO

from typing import Dict, Union


def test_parser_version():
    screen_stdout = sys.stdout
    string_stdout = StringIO()
    sys.stdout = string_stdout
    
    try:
        __main__.parser.parse_args(["--version"])
    except SystemExit:
        output = string_stdout.getvalue()
        expected = CytoOne.__version__.__version__ + "\n"
        assert output == expected
        sys.stdout = screen_stdout
    else:
        sys.stdout = screen_stdout
        assert False
        

@pytest.mark.parametrize("cmdarg", ["--help", "-h"])
def test_parser_help(cmdarg):
    screen_stdout = sys.stdout
    string_stdout = StringIO()
    sys.stdout = string_stdout
    
    try:
        __main__.parser.parse_args([cmdarg])
    except SystemExit:
        output = string_stdout.getvalue()
        assert "usage: " in output
        sys.stdout = screen_stdout
    else:
        sys.stdout = screen_stdout
        assert False







# Data IO 
import os 
import json
# Data manipulation
import numpy as np 
import pandas as pd 
from scipy.stats import ks_2samp
# PyTorch
import torch 
import torch.nn as nn 
import torch.nn.functional as F
from torch import optim
from torch.distributions import Independent, Normal
# Modules 
from CytoOne.encoder import Encoder
from CytoOne.decoder import Decoder
from CytoOne.utilities import import_data, generate_strata,\
                             load_stratum, reparameterize, \
                             kl_standard, mmd_loss, JSONEncoder
from CytoOne.basic_distributions import QuasiZeroInflatedSoftplusNormal, QuasiZeroInflatedLogNormal
# User entertainment
from tqdm.auto import tqdm 
from typing import Optional, Union, Tuple


class cytoone(nn.Module):
    def __init__(self,
                 batch_index_col: Optional[str]=None,
                 celltype_col: Optional[str]=None,
                 normalize: bool=True,
                 dr: bool=True, 
                 zero_inflated: bool=True,
                 latent_dims: list=[20, 10, 5, 2],
                 batch_embedding_dim: int=2, 
                 encoder_hidden_dims: list=[[1024, 512, 256], [512, 256, 128], [256, 128, 64], [128, 64,32]],
                 decoder_hidden_dims: list=[[32, 64,128], [64, 128, 256], [128, 256, 512], [256, 512, 1024]],
                 drop_out_p: float=0.2,
                 top_gamma: float=2.0,
                 top_beta: float=0.01,
                 pretrain_gamma: float=1.0,
                 pretrain_beta: float=1.0,
                #  anneal_percent: float=0.0,
                 distribution_type: str="softplus_normal",
                 model_device: Optional[Union[str, torch.device]] = None) -> None:
        """Initialize cytoone object 

        Parameters
        ----------
        batch_index_col : Optional[str], optional
            The column containing batch information, by default None
        celltype_col : Optional[str], optional
            The column containing cell type information, by default None
        normalize : bool, optional
            If normalize the data via asinh, by default True
        dr : bool, optional
            If UMAP should be computed, by default True
        zero_inflated : bool, optional
            If the data if zero-inflated, by default True
        latent_dims : list, optional
            In a bottom-up fashion, the dimensions of latent variables, by default [20, 10, 5, 2]
        batch_embedding_dim : int, optional
            Dimension of batch embeddings, by default 2
        encoder_hidden_dims : list, optional
            A nested list where each sublist contains the number of hidden units for the encoder, by default [[1024, 512, 256], [512, 256, 128], [256, 128, 64], [128, 64,32]]
        decoder_hidden_dims : list, optional
            A nested list where each sublist contains the number of hidden units for the decoder, by default [[32, 64,128], [64, 128, 256], [128, 256, 512], [256, 512, 1024]]
        drop_out_p : float, optional
            Drop out probability, by default 0.2
        top_gamma : float, optional
            Top level penalty for MMD, by default 2.0
        top_beta : float, optional
            Top level penalty for KL divergence, by default 0.01
        pretrain_gamma : float, optional
            Top level penalty for MMD during pretrain, by default 1.0
        pretrain_beta : float, optional
            Top level penalty for KL divergence during pretrain, by default 1.0
        distribution_type : str, optional
            Likelihood to use, by default "softplus_normal"
        model_device : Optional[Union[str, torch.device]], optional
            Model device, by default None
        """
        super().__init__()
        # Parameters for importing data 
        self.import_data_par = {"batch_index_col": batch_index_col,
                                "celltype_col": celltype_col,
                                "normalize": normalize,
                                "dr": dr,
                                "zero_inflated": zero_inflated}
        # Parameters for initializing the encoder
        self.encoder_par = {"input_dim": None,
                            "batch_embedding_dim": batch_embedding_dim, 
                            "latent_dims": latent_dims,
                            "hidden_dims": encoder_hidden_dims,
                            "drop_out_p": drop_out_p}
        # Parameters for initializing the decoder
        self.decoder_par = {"input_dim": None,
                            "batch_embedding_dim": batch_embedding_dim,  
                            "latent_dims": latent_dims[::-1],
                            "hidden_dims": decoder_hidden_dims,
                            "drop_out_p": drop_out_p}
        # Data
        self.adata = None
        self.n_batches = None
        self.zero_inflated = zero_inflated
        self.distribution_type = distribution_type
        # Set model device
        if model_device is None:
            self.model_device = torch.device(
                'cuda' if torch.cuda.is_available() else 'cpu')
        elif isinstance(model_device, str):
            self.model_device = torch.device(model_device)
        else:
            self.model_device = model_device
        # Main modules 
        self.encoder = None
        self.decoder = None
        self.rough_log_var = None
        self.noise_log_normal_var = None
        self.batch_embedding = None
        # Optimization 
        self.optimizer = None 
        # self.anneal_percent=anneal_percent
        self.top_beta = top_beta
        self.pretrain_beta = pretrain_beta
        self.beta = []
        self.top_gamma = top_gamma
        self.pretrain_gamma = pretrain_gamma
        self.gamma = []
        # if self.anneal_percent <= 0.0:
        #     self.beta = 1.0
        # else:
        #     self.beta = 0.0
        self.log_interval = 10
        # Monitoring loss 
        self.RECON_list = []
        self.KLD_list = []
        self.MMD_list = []

    def import_data(self,
                    cell_by_gene: Union[str, pd.DataFrame],
                    cell_metadata: Optional[Union[str, pd.DataFrame]]=None) -> None:
        """Import data 

        Parameters
        ----------
        cell_by_gene : Union[str, pd.DataFrame]
            Either the path to the csv file or pd.DataFrame that contains the CyTOF measurements 
        cell_metadata : Optional[Union[str, pd.DataFrame]], optional
            Either the path to the csv file or pd.DataFrame that contains metainformation of the CyTOF measurements , by default None
        """
        self.adata = import_data(cell_by_gene=cell_by_gene,
                                 cell_metadata=cell_metadata,
                                 **self.import_data_par)
        
        self.encoder_par['input_dim'] = self.adata.uns["n_genes"]
        self.decoder_par['input_dim'] = self.adata.uns["n_genes"]
        self.n_batches = self.adata.uns['n_batches']
        if not self.zero_inflated:
            neg_x = self.adata.X[self.adata.X<=0].copy().reshape(-1)
            self.rough_log_var = np.log(np.var(np.concatenate((neg_x, -neg_x))))

    def initialize_parameters(self):
        """Initialize parameters
        """
        self.encoder = Encoder(**self.encoder_par)
        self.decoder = Decoder(**self.decoder_par)

        self.batch_embedding = nn.Embedding(self.n_batches, self.encoder_par['batch_embedding_dim'])
        if not self.zero_inflated:
            self.noise_log_normal_var = nn.Parameter(self.rough_log_var*torch.ones(1), requires_grad=True)
            self.optimizer = optim.Adam([{'params': self.encoder.parameters()},
                                        {'params': self.decoder.parameters()},
                                        {'params': self.batch_embedding.parameters()},
                                        {'params': self.noise_log_normal_var}], lr=1e-3)
        else: 
            self.optimizer = optim.Adam([{'params': self.encoder.parameters()},
                                        {'params': self.decoder.parameters()},
                                        {'params': self.batch_embedding.parameters()}], lr=1e-3)
        self.to(self.model_device)

    def encode(self,
               cell_by_gene_counts: torch.tensor,
               source_batch_index: torch.tensor,
               mode: str="random",
               pretrain: bool=False) -> Tuple[torch.tensor, torch.tensor, list, torch.tensor]:
        """Encode the observed measurement 

        Parameters
        ----------
        cell_by_gene_counts : torch.tensor
            The CyTOF measurement
        source_batch_index : torch.tensor
            The associated batch information 
        mode : str, optional
            If sampling is enabled, by default "random"
        pretrain : bool, optional
            Use the pretain model , by default False

        Returns
        -------
        Tuple[torch.tensor, torch.tensor, list, torch.tensor]
            Location and log-var of the top-level latent. A list of intermediate latent. Top-level z 
        """
        # Encoder will generate the mu and log_var of the top-level z
        # xs is a list of output of residule blocks
        mu, log_var, xs = self.encoder(x=cell_by_gene_counts,
                                        batch_index=source_batch_index,
                                        batch_embedding=self.batch_embedding,
                                        pretrain=pretrain)
        if mode=='fix':
            z = reparameterize(mu, 0)
        else:
            # Randomly sample top-level z 
            z = reparameterize(mu, torch.exp(0.5 * log_var))
        return mu, log_var, xs, z

    def decode(self,
               z: torch.tensor,
               target_batch_index: torch.tensor,
               xs: list,
               mode: str='random',
               denoise: bool=False,
               pretrain: bool=False) -> Tuple[torch.distributions.Distribution, list, list]:
        """Decode the latent variables 

        Parameters
        ----------
        z : torch.tensor
            Top-level latent variable 
        target_batch_index : torch.tensor
            Batch to which batch correction is targeted 
        xs : list
            A list of intermediate latent 
        mode : str, optional
            If sampling is enabled , by default 'random'
        denoise : bool, optional
            If the output should be zero-inflated, by default False
        pretrain : bool, optional
            Whether to use pretrain model, by default False

        Returns
        -------
        Tuple[torch.distributions.Distribution, list, list]
            Distribution of the measuement. KL losses and intermediate latent variables 

        """
        # Based on the zero inflated, we use different likelihood 
        x_mu, x_log_var, x_gate_logit, kl_losses, zs = self.decoder(z=z,
                                                                batch_index=target_batch_index,
                                                                batch_embedding=self.batch_embedding,
                                                                xs=xs,
                                                                mode=mode,
                                                                pretrain=pretrain) 
        if denoise or self.zero_inflated or (self.distribution_type=='normal'):
            normal_scale = None
        else:
            normal_scale = torch.exp(0.5*self.noise_log_normal_var)
        
        x_scale = torch.exp(0.5*x_log_var)
        if self.distribution_type == "softplus_normal":
            x_dists = Independent(QuasiZeroInflatedSoftplusNormal(loc=x_mu,
                                                scale=x_scale,
                                                gate_logits=x_gate_logit,
                                                normal_scale=normal_scale), 0)

        elif self.distribution_type == "log_normal":
            x_dists = Independent(QuasiZeroInflatedLogNormal(loc=x_mu,
                                                scale=x_scale,
                                                gate_logits=x_gate_logit,
                                                normal_scale=normal_scale), 0)
        elif self.distribution_type == "normal":
            x_dists = Independent(Normal(loc=x_mu, scale=x_scale), 0)
        else: 
            raise TypeError("Unknown distribution type")

        return x_dists, kl_losses, zs

    def forward(self,
                cell_by_gene_counts: torch.tensor,
                source_batch_index: torch.tensor,
                target_batch_index: torch.tensor,
                pretrain: bool=False) -> Tuple[torch.distributions.Distribution, list, list]:
        """The forward pass 

        Parameters
        ----------
        cell_by_gene_counts : torch.tensor
            The CyTOF measurements 
        source_batch_index : torch.tensor
            The associated batch information 
        target_batch_index : torch.tensor
            The batch to which batch correction is targetd 
        pretrain : bool, optional
            Whether to use pretrain model, by default False

        Returns
        -------
        Tuple[torch.distributions.Distribution, list, list]
            Distribution of the measuement. KL losses and intermediate latent variables 
        """
        mu, log_var, xs, z = self.encode(cell_by_gene_counts=cell_by_gene_counts,
                                         source_batch_index=source_batch_index,
                                         pretrain=pretrain) 
        x_dists, kl_losses, zs = self.decode(z=z,
                                             target_batch_index=target_batch_index,
                                             xs=xs,
                                             pretrain=pretrain)
        # KL divergence top-level 
        kl_losses = [kl_standard(mu=mu, log_var=log_var)] + kl_losses
        # kl_losses.append(kl_standard(mu=mu, log_var=log_var)) 
        return x_dists, kl_losses, zs

    def infer(self,
              new_cell_by_gene: Optional[Union[str, pd.DataFrame]]=None,
              new_cell_metadata: Optional[Union[str, pd.DataFrame]]=None,
              target_batch_index: Optional[int]=None,
              mode: str='random',
              denoise: bool=False,
              use_pretrain: bool=False,
              get_normal_component: bool=False) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """Perform downstream analysis 

        Parameters
        ----------
        new_cell_by_gene : Optional[Union[str, pd.DataFrame]], optional
            New CyTOF measurements, by default None
        new_cell_metadata : Optional[Union[str, pd.DataFrame]], optional
            New associated meta information, by default None
        target_batch_index : Optional[int], optional
            Batch to which batch correction is targeted, by default None
        mode : str, optional
            If sampling is enabled, by default 'random'
        denoise : bool, optional
            Whether zero-inflated measurement should be generated, by default False
        use_pretrain : bool, optional
            Whether to use pretrain model, by default False
        get_normal_component : bool, optional
            Whether to get the normal component, by default False

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            Generated CyTOF measurement, Top-level latent variables 
        """
        self.eval()
        with torch.no_grad():
            if new_cell_by_gene is None:
                adata_w_batch = self.adata.to_df().copy()
                adata_w_batch['batch_index'] = self.adata.obs['batch_index'].copy() 
            else: 
                new_adata = import_data(cell_by_gene=new_cell_by_gene,
                                        cell_metadata=new_cell_metadata,
                                        **self.import_data_par)
                adata_w_batch = new_adata.to_df().copy()
                adata_w_batch['batch_index'] = new_adata.obs['batch_index'].copy() 
            splits = np.array_split(adata_w_batch.index, 100)
            x_samples = []
            z_samples = []
            for i, row_ind in enumerate(splits):
                source_batch_index = torch.tensor(adata_w_batch.loc[row_ind, "batch_index"].values, 
                                                  dtype=torch.int64, device=self.model_device)
                
                cell_by_gene_counts = torch.tensor(adata_w_batch.loc[row_ind, :].drop(columns=['batch_index']).values,
                                                   dtype=torch.float32, device=self.model_device)
                
                _, _, xs, z = self.encode(cell_by_gene_counts=cell_by_gene_counts,
                                                source_batch_index=source_batch_index,
                                                mode=mode,
                                                pretrain=use_pretrain) 
                if target_batch_index is None:
                    t_ind = source_batch_index.clone()
                else:
                    t_ind = torch.ones_like(source_batch_index) * target_batch_index
                x_dists, _, _ = self.decode(z=z,
                                            target_batch_index=t_ind,
                                            xs=xs,
                                            mode=mode,
                                            denoise=denoise,
                                            pretrain=use_pretrain)

                if get_normal_component and ("Quasi" in str(type(x_dists.base_dist))):
                    x_samples.append(x_dists.base_dist.base_dist.base_dist.sample().detach().cpu().numpy()) 
                else:
                    x_samples.append(x_dists.sample().detach().cpu().numpy())
                z_samples.append(z.detach().cpu().numpy())
            
            x_samples = pd.DataFrame(np.concatenate(x_samples, axis=0))
            x_samples.columns = self.adata.var_names
            x_samples.index = adata_w_batch.index
            x_samples['source_batch_index'] = adata_w_batch['batch_index'].copy() 
            if target_batch_index is None:
                x_samples['batch_index'] = adata_w_batch['batch_index'].copy() 
            else:
                x_samples['batch_index'] = target_batch_index
            z_samples = pd.DataFrame(np.concatenate(z_samples, axis=0),
                                     columns=["z"+str(i) for i in range(self.encoder_par['latent_dims'][-1])])
            z_samples.index = self.adata.obs_names
            return x_samples, z_samples
                 
    def loss_function(self,
                      x_dists: torch.distributions.Distribution,
                      cell_by_gene_counts: torch.tensor,
                      kl_losses: list,
                      zs: list,
                      source_batch_index: torch.tensor) -> Tuple[torch.tensor, float, float, float]:
        """Compute the loss 

        Parameters
        ----------
        x_dists : torch.distributions.Distribution
            Distribution of the measuement
        cell_by_gene_counts : torch.tensor
            The CyTOF measurements 
        kl_losses : list
            KL losses
        zs : list
            Intermediate latent variables 
        source_batch_index : torch.tensor
            The associated batch information 

        Returns
        -------
        Tuple[torch.tensor, float, float, float]
            -ELBO, log likelihood, KLD, MMD 
        """
        # Likelihood 
        log_likelihood = x_dists.log_prob(cell_by_gene_counts).sum(dim=1).mean()

        KLD = 0.0
        for i, k in enumerate(kl_losses[::-1]):
            KLD += k * self.beta[i]
        
        # MMD 
        MMD = 0.0
        if self.n_batches > 1:
            for i, z in enumerate(zs[::-1]):
                batch_l = mmd_loss(z=z, batch_index=source_batch_index)
                MMD += batch_l * self.gamma[i]
        else:
            MMD = torch.zeros(1, dtype=torch.float32)

        return -log_likelihood + KLD + MMD,\
                log_likelihood.detach().cpu().numpy().item(),\
                KLD.detach().cpu().numpy().item(),\
                MMD.detach().cpu().numpy().item()

    def _training_loop(self,
                      n_epoches: int,
                      n_strata: int,
                      early_stop_pval: float,
                      pretrain: bool) -> None:
        """Model training loop 

        Parameters
        ----------
        n_epoches : int, optional
            Number of epoches
        n_strata : int, optional
            Number of minibatches per epoch
        early_stop_pval : float, optional
            p-value used to detect early stopping
        pretrain : bool, optional
            Whether or not the training is pretraining
        """
        # total_anneal_steps = np.round(n_epoches * n_strata * self.anneal_percent)
        if not pretrain:
            # If not pretraining, we freeze pretrain models 
            for param in self.encoder.encoder_elevator.parameters():
                param.requires_grad = False
            for param in self.decoder.decoder_elevator.parameters():
                param.requires_grad = False
            # Compute penalties for KLD and MMD 
            latent_dims = self.encoder_par['latent_dims']
            self.beta = [i/np.max(latent_dims) for i in latent_dims]
            self.beta[-1] = np.minimum(self.beta[-1], self.top_beta)
            self.gamma = [self.top_gamma*np.min(latent_dims)/i for i in latent_dims]
        else: 
            self.gamma = [self.pretrain_gamma]
            self.beta = [self.pretrain_beta]
        self.train()
        for epoch in range(n_epoches):
            # For epoch, we randomly suffule the data and stratify 
            adata_w_batch_strata = generate_strata(adata=self.adata,
                                                   n_strata=n_strata)
            RECON_epoch_list = []
            KLD_epoch_list = []
            MMD_epoch_list = []
            # Starting from the 3rd epoch, we test if convergence has been achieved 
            if epoch >= 2:
                RECON_previous_2 = np.array(self.RECON_list[epoch-2])
                RECON_previous_1 = np.array(self.RECON_list[epoch-1])
                p_val = ks_2samp(RECON_previous_2, RECON_previous_1).pvalue
                if p_val > early_stop_pval:
                    print("="*30)
                    print("No improvement in the reconstruction task detected. Stop early at epoch {}".format(epoch-1))
                    print("="*30)
                    break
            train_loss = 0.0
            for minibatch_ind in range(n_strata):
                cell_by_gene_counts, source_batch_index, target_batch_index = load_stratum(adata_w_batch_strata=adata_w_batch_strata,
                                                                                           target_batch_index=None,
                                                                                           stratum_id=minibatch_ind,
                                                                                           model_device=self.model_device)
                x_dists, kl_losses, zs = self(cell_by_gene_counts=cell_by_gene_counts,
                                              source_batch_index=source_batch_index,
                                              target_batch_index=target_batch_index,
                                              pretrain=pretrain)
                
                self.optimizer.zero_grad()
                loss, RECON, KLD, MMD = self.loss_function(x_dists=x_dists,
                                                    cell_by_gene_counts=cell_by_gene_counts,
                                                    kl_losses=kl_losses,
                                                    zs=zs,
                                                    source_batch_index=source_batch_index)
                loss.backward()
                RECON_epoch_list.append(RECON)
                KLD_epoch_list.append(KLD)
                MMD_epoch_list.append(MMD)
                train_loss += loss.item()
                self.optimizer.step()
                # # KL Annealing 
                # if self.anneal_percent > 0:
                #     self.beta = np.minimum(1.0, (n_strata*epoch+minibatch_ind)/total_anneal_steps)

                # Print training information 
                if minibatch_ind % self.log_interval == 0:
                    print('Train Epoch: {} [{}/{} ({:.0f}%)]tLoss: {:.6f}'.format(
                            epoch, minibatch_ind, n_strata,
                                100. * minibatch_ind / n_strata,
                                loss.item()))
            print('====> Epoch: {} Average loss: {:.4f}'.format(
                    epoch, train_loss / n_strata))
            self.RECON_list.append(np.array(RECON_epoch_list))
            self.KLD_list.append(np.array(KLD_epoch_list))
            self.MMD_list.append(np.array(MMD_epoch_list))
    
    def training_loop(self,
                      n_epoches: int=50,
                      n_strata: int=100,
                      early_stop_pval: float=1.0) -> None:
        """Model training loop

        Parameters
        ----------
        n_epoches : int, optional
            Number of epoches, by default 50
        n_strata : int, optional
            Number of minibatches per epoch, by default 100
        early_stop_pval : float, optional
            p-value used to detect early stopping, by default 1.0
        """
        self._training_loop(n_epoches=n_epoches,
                            n_strata=n_strata,
                            early_stop_pval=early_stop_pval,
                            pretrain=True)
        self._training_loop(n_epoches=n_epoches,
                            n_strata=n_strata,
                            early_stop_pval=early_stop_pval,
                            pretrain=False)
        
    def save_model(self,
                   dir_name: str,
                   model_name: str="cytoone") -> None:
        """Save model results

        Parameters
        ----------
        dir_name : str
            The directory path at which the saved model should be.
        model_name : str
            The model name, by default cytoone
        """
        torch.save({'model_state_dict': self.state_dict()} | \
                {'optimizer_state_dict': self.optimizer.state_dict()}, 
                os.path.join(dir_name, model_name+".pt")) 

        model_meta = {"import_data_par": self.import_data_par,
                      "encoder_par": self.encoder_par,
                      "decoder_par": self.decoder_par,
                      "n_batches": self.n_batches,
                      "zero_inflated": self.zero_inflated,
                      "rough_log_var": self.rough_log_var}
        with open(os.path.join(dir_name, model_name+"_meta.json"), "w") as f:
            json.dump(model_meta, f, cls=JSONEncoder)
        
    def load_model(self,
                   dir_name: str,
                   model_name: str) -> None:
        """Load model results 

        Parameters
        ----------
        dir_name : str
            The directory path at which the saved model should be.
        model_name : str
            The model name
        """
        model_meta = json.load(open(os.path.join(dir_name, model_name+"_meta.json")))
        self.import_data_par = model_meta['import_data_par']
        self.encoder_par = model_meta['encoder_par']
        self.decoder_par = model_meta['decoder_par']
        self.n_batches = model_meta['n_batches']
        self.zero_inflated = model_meta['zero_inflated']
        self.rough_log_var = model_meta["rough_log_var"]
        
        self.initialize_parameters()
        checkpoint = torch.load(os.path.join(dir_name, model_name+".pt")) 
        self.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])

    
    


__title__ = "CytoOne"
__description__ = "A unified probabilistic framework for CyTOF data"
__version__ = "0.0.2"
__author__ = "Yuqiu Yang; Kaiwen Wang; Xinlei Wang"
__authoremail__ = "yuqiu.yang@utsouthwestern.edu; @.edu; @.edu"


from setuptools import setup 
import re 
import os 


VERSIONFILE="CytoOne/__version__.py"
verstrline = open(VERSIONFILE, "rt").read()
# Version 
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    VERSION = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))
# Authors
AURE = r"^__author__ = ['\"]([^'\"]*)['\"]"
mo = re.search(AURE, verstrline, re.M)
if mo:
    AUTHORS = mo.group(1)
else:
    raise RuntimeError("Unable to find author string in %s." % (VERSIONFILE,))
# Description
DESRE = r"^__description__ = ['\"]([^'\"]*)['\"]"
mo = re.search(DESRE, verstrline, re.M)
if mo:
    DESCRIPTION = mo.group(1)
else:
    raise RuntimeError("Unable to find description string in %s." % (VERSIONFILE,))
# Title 
TITRE = r"^__title__ = ['\"]([^'\"]*)['\"]"
mo = re.search(TITRE, verstrline, re.M)
if mo:
    TITLE = mo.group(1)
else:
    raise RuntimeError("Unable to find title string in %s." % (VERSIONFILE,))

# Title 
AUEMRE = r"^__authoremail__ = ['\"]([^'\"]*)['\"]"
mo = re.search(AUEMRE, verstrline, re.M)
if mo:
    AUTHOREMAIL = mo.group(1)
else:
    raise RuntimeError("Unable to find author email string in %s." % (VERSIONFILE,))


lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = f"{lib_folder}/requirements.txt"
install_requires = [] 
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()


setup(
    name=TITLE,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHORS,
    author_email=AUTHOREMAIL,
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    packages=["CytoOne"],
    python_requires=">=3.9,<3.11",
    install_requires=install_requires,
    test_requires=[
        'pytest==7.1.2',
        'coverage==6.3.2',
        'pytest-cov==3.0.0',
        'pytest-mock==3.10.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Natural Language :: English",
        "Intended Audience :: Science/Research",
    ]
)